@@ -247,6 +247,10 @@ master_get_table_ddl_events(PG_FUNCTION_ARGS)
  * Please note that the caller is still responsible for finalizing shard data
  * and the shardId with the master node. Further note that this function relies
  * on an internal sequence created in initdb to generate unique identifiers.
+ *
+ * NB: This can be called by any user; for now we have decided that that's
+ * ok. We might want to restrict this to users part of a specific role or such
+ * at some later point.
  */
 Datum
 master_get_new_shardid(PG_FUNCTION_ARGS)
@@ -254,12 +258,19 @@ master_get_new_shardid(PG_FUNCTION_ARGS)
 	text *sequenceName = cstring_to_text(SHARDID_SEQUENCE_NAME);
 	Oid sequenceId = ResolveRelationId(sequenceName);
 	Datum sequenceIdDatum = ObjectIdGetDatum(sequenceId);
+	Oid savedUserId = InvalidOid;
+	int savedSecurityContext = 0;
+	Datum shardIdDatum = 0;
+
+	GetUserIdAndSecContext(&savedUserId, &savedSecurityContext);
+	SetUserIdAndSecContext(CitusExtensionOwner(), SECURITY_LOCAL_USERID_CHANGE);
 
 	/* generate new and unique shardId from sequence */
-	Datum shardIdDatum = DirectFunctionCall1(nextval_oid, sequenceIdDatum);
-	int64 shardId = DatumGetInt64(shardIdDatum);
+	shardIdDatum = DirectFunctionCall1(nextval_oid, sequenceIdDatum);
 
-	PG_RETURN_INT64(shardId);
+	SetUserIdAndSecContext(savedUserId, savedSecurityContext);
+
+	PG_RETURN_DATUM(shardIdDatum);
 }
 
 
