@@ -1169,7 +1169,7 @@ MultiSelectNode(List *whereClauseList)
 /*
  * IsSelectClause determines if the given node is a select clause according to
  * our criteria. Our criteria defines a select clause as an expression that has
- * columns belonging to only one table.
+ * zero or more columns belonging to only one table.
  */
 static bool
 IsSelectClause(Node *clause)
@@ -1179,20 +1179,22 @@ IsSelectClause(Node *clause)
 	Var *firstColumn = NULL;
 	Index firstColumnTableId = 0;
 	bool isSelectClause = true;
-
-	/* we currently consider the following nodes as select clauses */
 	NodeTag nodeTag = nodeTag(clause);
-	if (!(nodeTag == T_OpExpr || nodeTag == T_ScalarArrayOpExpr ||
-		  nodeTag == T_NullTest || nodeTag == T_BooleanTest))
+
+	/* error out for subqueries in WHERE clause */
+	if (nodeTag == T_SubLink || nodeTag == T_SubPlan)
 	{
-		return false;
+		ereport(ERROR, (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+						errmsg("cannot perform distributed planning on this query"),
+						errdetail("Subqueries other than in from-clause are currently "
+								  "unsupported")));
 	}
 
 	/* extract columns from the clause */
 	columnList = pull_var_clause_default(clause);
 	if (list_length(columnList) == 0)
 	{
-		return false;
+		return true;
 	}
 
 	/* get first column's tableId */
@@ -1558,7 +1560,7 @@ ApplyJoinRule(MultiNode *leftNode, MultiNode *rightNode, JoinRuleType ruleType,
 	List *applicableJoinClauses = NIL;
 	List *leftTableIdList = OutputTableIdList(leftNode);
 	List *rightTableIdList = OutputTableIdList(rightNode);
-	int rightTableIdCount = 0;
+	int rightTableIdCount PG_USED_FOR_ASSERTS_ONLY = 0;
 	uint32 rightTableId = 0;
 
 	rightTableIdCount = list_length(rightTableIdList);
