@@ -18,6 +18,9 @@ extern PlannedStmt * multi_planner(Query *parse, int cursorOptions,
 
 extern bool HasCitusToplevelNode(PlannedStmt *planStatement);
 struct MultiPlan;
+extern struct MultiPlan * CreatePhysicalPlan(Query *parse);
 extern struct MultiPlan * GetMultiPlan(PlannedStmt *planStatement);
+extern PlannedStmt * MultiQueryContainerNode(PlannedStmt *result,
+											 struct MultiPlan *multiPlan);
 
 #endif /* MULTI_PLANNER_H */
